/**
 * Get the HTML for a currency selector select element
 *
 * @param String dom_id
 * @param String name
 * @param String selected_value
 *
 * @return String
 */
function wc_vat_compliance_currency_selector_dropdown(dom_id, name, selected_value) {
	
	html = '<select id="'+dom_id+'" name="'+name+'" class="wc_euvat_currency_selector">';
	
	var currencies = wc_vat_compliance.currency_list;
	
	for (const code in currencies) {
		var selected = (code == selected_value) ? ' selected="selected"' : '';
		html += '<option value="'+code+'"'+selected+'>'+currencies[code]+'</option>';
	}
	
	html += '</select>'+"\n";
	
	return html;
}

jQuery(function($) {
	
	var value_exemption_selector = wc_vat_compliance_currency_selector_dropdown('woocommerce_eu_vat_cart_vat_exempt_above_currency', 'woocommerce_eu_vat_cart_vat_exempt_above_currency', wc_vat_compliance.selected_recording_currency);
	$(value_exemption_selector).insertAfter('#woocommerce_eu_vat_cart_vat_exempt_above');
	
	var vat_number_override_index = 0;
	
	/**
	 * Get the HTML for a table row for deleting the row
	 *
	 * @return String
	 */
	function delete_row() {
		var html = '<a href="#" class="wcvat_delete_table_row">'+wc_vat_compliance.delete_this_override+"\n";
		return html;
	}
	
	$('#wcvat-number-entry-overrides').on('click', '.wcvat_delete_table_row', function() {
		$(this).parents('.wcvat-number-entry-override').first().slideUp(function() { $(this).remove(); });
		return false;
	});
	
	/**
	 * Return the HTML for a region selector dropdown
	 *
	 * @param String dom_id
	 * @param String name
	 * @param String selected_value
	 *
	 * @return String
	 */
	function region_selector_dropdown(dom_id, name, selected_value) {
		
		html = '<select id="'+dom_id+'" name="'+name+'" class="wc_vat_region_selector">';
		
		for (const region_code in wc_vat_compliance.region_list) {
			var selected = (region_code == selected_value) ? ' selected="selected"' : '';
			html += '<option value="'+region_code+'"'+selected+'>'+wc_vat_compliance.region_list[region_code]+'</option>';
		}
		
		html += '</select>'+"\n";
		
		return html;
	}
	
	/**
	 * Get the HTML for a region selector widget
	 *
	 * @param Integer index
	 * @param String  selected_value
	 *
	 * @return String
	 */
	function region_selector_widget(index, selected_value) {
		
		index = index.toString();
		
		var dom_id = 'wc_vat_region_selector_'+index;
		
		var html = wc_vat_compliance.in_region_use_policy.replace('%s', '<span class="forminp forminp-select">'+region_selector_dropdown(dom_id, 'woocommerce_eu_vat_number_entry_overrides['+index+'][region]', selected_value)+'</span>')+"\n";
		
		return html;
		
	}
	
	/**
	 * Return the HTML for a VAT number policy selector dropdown
	 *
	 * @param String dom_id
	 * @param String name
	 * @param String selected_value
	 *
	 * @return String
	 */
	function vat_number_entry_policy_selector_dropdown(dom_id, name, selected_value) {
		
		html = '<select id="'+dom_id+'" name="'+name+'" class="wc_vat_vat_number_policy_selector">';
		
		for (const policy_id in wc_vat_compliance.vat_number_policies) {
			var selected = (policy_id == selected_value) ? ' selected="selected"' : '';
			html += '<option value="'+policy_id+'"'+selected+'>'+wc_vat_compliance.vat_number_policies[policy_id]+'</option>';
		}
		
		html += '</select>'+"\n";
		
		return html;
	}
	
	/**
	 * Get the HTML for a VAT number policy selector widget
	 *
	 * @param Integer index
	 * @param String  selected_value
	 *
	 * @return String
	 */
	function region_vat_number_entry_policy_widget(index, selected_value) {
		
		index = index.toString();
		
		var dom_id = 'wc_vat_region_selector_'+index;
		
		var html = '<span class="forminp forminp-select">'+"\n";
		
		html += vat_number_entry_policy_selector_dropdown(dom_id, 'woocommerce_eu_vat_number_entry_overrides['+index+'][policy]', selected_value);
		
		html += "</span>\n";
		
		return html;
		
	}
	
	$('#wcvat-new-number-entry-override').on('click', function() {
		
		vat_number_override_index++;
		
		var new_number_override = '<div class="wcvat-number-entry-override">' + region_selector_widget(vat_number_override_index) + region_vat_number_entry_policy_widget(vat_number_override_index) + delete_row() + '</div>';
		
		$('#wcvat-number-entry-overrides').append(new_number_override);
		
		return false;
	});
	
	var existing_overrides = $('#wcvat-new-number-entry-override').data('existing-overrides');
	
	if ('object' == typeof existing_overrides) {
		for (const region_code in existing_overrides) {
			
			var new_number_override = '<div class="wcvat-number-entry-override">' + region_selector_widget(vat_number_override_index, region_code) + region_vat_number_entry_policy_widget(vat_number_override_index, existing_overrides[region_code]) + delete_row() + '</div>';
			
			$('#wcvat-number-entry-overrides').append(new_number_override);
			
			vat_number_override_index++;
		}
		
		
	}
	
	
});
